/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.mim.lib;

import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.RolledBackPersistenceException;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.speedo.api.Debug;
import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.metadata.SpeedoFetchGroup;
import org.objectweb.speedo.mim.api.DetachedLifeCycle;
import org.objectweb.speedo.mim.api.SpeedoHome;
import org.objectweb.speedo.mim.api.SpeedoAccessor;
import org.objectweb.speedo.mim.api.SpeedoProxy;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.speedo.pm.api.ProxyManagerFactory;
import org.objectweb.speedo.query.api.QueryDefinition;
import org.objectweb.speedo.usercache.api.UserCache;
import org.objectweb.speedo.usercache.lib.UserCacheImpl;
import org.objectweb.speedo.workingset.api.Transaction;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.jdo.JDOException;
import javax.jdo.JDOUserException;

/**
 * @author S.Chassande-Barrioz
 */
public abstract class SpeedoHomeImpl implements SpeedoHome {

    protected TransactionalPersistenceManager tpm = null;

	protected ProxyManagerFactory pmf = null;
	
	byte cachePolicy = CACHED;
	
	boolean isFieldLockingLevel = false;
	boolean prefetchOnQuery = true;
	boolean prefetchOnExtent = true;
	boolean prefetchOnGenClass = true;
	
	protected Map namedQueries = null;
	
	public SpeedoHomeImpl() {
	}
	
	public SpeedoHomeImpl(TransactionalPersistenceManager _tpm, 
	        ProxyManagerFactory _pmf) {
	    this.tpm = _tpm;
	    this.pmf = _pmf;
	}
	
	/**
	 * @return false because it corresponds to most of case
	 * Subclass to change the behavior
	 */
	protected boolean isAbstract() {
	    return false;
	}
	
    /**
     * Create a new instance of a SpeedoProxy which the clas is given in 
     * parameter. The default implementation does a 
     * simple 'clazz.newInstance()'.
     */
	protected SpeedoProxy newSpeedoProxyInstance(Class clazz) throws Exception {
        return (SpeedoProxy) clazz.newInstance();
    }

	// IMPLEMENTATION OF THE ProxyHome INTERFACE //
	//-------------------------------------------//
	
	public boolean isCacheable() {
	    return (cachePolicy & CACHED) != 0;
	}

    public boolean allLoaded() {
        return (cachePolicy & ALL) != 0;
    }
    
    public boolean hasToFix() {
        return (cachePolicy & FIXED) != 0;
    }
    
	public void setCachePolicy(byte v) {
	    cachePolicy = v;
	}

    public boolean isFieldLockingLevel() {
        return isFieldLockingLevel;
    }

    public void setFieldLockingLevel(boolean val) {
        isFieldLockingLevel = val;
    }
    
    /**
	 * This default implementation does nothing and returns null.
	 * Subclass to change the behavior
	 */
    public Collection fgGetNestedFetchGroups(String fgName) {
	    return null;
	}

	/**
	 * This default implementation returns always false.
	 * Subclass to change the behavior
	 */
	public boolean fgIsDefined(String fgName) {
	    return false;
	}

	/**
	 * This default implementation does nothing and returns null.
	 * Subclass to change the behavior
	 */
	public Collection fgGetFieldsToLoad(String fgName) {
	    return null;
	}
	
    /**
	 * @return the PersistenceManagerFactory which represents the data support
	 * inside which the proxy is persistent.
	 */
	public final ProxyManagerFactory getProxyManagerFactory() {
		return pmf;
	}

	/**
	 * It assignes a PersistenceManagerFactory
	 * @param _pmf is the PersistenceManagerFactory which represents the data
	 * support inside which the proxy is persistent.
	 */
	public final void setProxyManagerFactory(ProxyManagerFactory _pmf) {
		this.pmf = _pmf;
	}


	/**
	 * @return the TransactionalPersistenceManager which manage the
	 * concurrency, loading, ... of the proxy.
	 */
	public final TransactionalPersistenceManager getTransactionalPersistenceManager() {
		return tpm;
	}

	/**
	 * It assignes a TransactionalPersistenceManager
	 * @param _tpm is the TransactionalPersistenceManager which manage the
	 * concurrency, loading, ... of the proxy.
	 */
	public final void setTransactionalPersistenceManager(TransactionalPersistenceManager _tpm) {
		this.tpm = _tpm;
	}

	/**
	 * Notifies the transactional persistency manager of a read intention for
	 * the given persistent instance.
	 * @param sp is the SpeedoProxy which the read access is requested.
	 * @param fields the ids of the fields that may be accessed by the caller
	 * of this method. If the i-th bit of 'fields' is set to 1, then the i-th
	 * field of the given speedo accessor may be accessed by the caller of this
	 * method.
	 */
	public final SpeedoAccessor readIntention(SpeedoProxy sp, long[] fields) {
		if (!sp.jdoIsActive()) {
			SpeedoAccessor sa = sp.getReferenceAccessor();
			if (sa == null) {
				sp.setReferenceAccessor(sp.createAccessor());
			}
			return sa;
		}
		ProxyManager pm = pmf.lookup();
		if (pm == null) {
		    throw new JDOUserException("When a persistent object is used (read), a PersistenceManager is needed");
		}
		Transaction t = (Transaction) pm.currentTransaction();
		try {
		    SpeedoAccessor sa = (SpeedoAccessor) tpm.readIntention(t, sp, null);
		    sa.loadFields(pm, fields);
		    return sa;
        } catch (RolledBackPersistenceException e) {
            throw t.rollBackOnInternalError(e);
		} catch (PersistenceException e) {
			throw new JDOException ("Impossible to notify a read intention", 
			        ExceptionHelper.getNested(e));
		}
	}

	public final SpeedoAccessor writeIntention(SpeedoProxy sp, long[] fields) {
	    return writeIntention(sp, fields, null);
	}
	
	/**
	 * Notifies the transactional persistency manager of a write intention for
	 * the given persistent instance.
	 *
	 * @param sp is the SpeedoProxy which the write access is requested.
	 * @param fields the ids of the fields that may be accessed by the caller
	 * of this method. If the i-th bit of 'fields' is set to 1, then the i-th
	 * field of the given speedo accessor may be accessed by the caller of this
	 * method.
	 */
	public SpeedoAccessor writeIntention(SpeedoProxy sp, long[] fields, Object thinLock) {
		if (!sp.jdoIsActive()) {
			SpeedoAccessor sa = sp.getReferenceAccessor();
			if (sa == null) {
				sp.setReferenceAccessor(sp.createAccessor());
			}
			return sa;
		}
		ProxyManager pm = pmf.lookup();
		if (pm == null) {
		    throw new JDOUserException("When a persistent object is used (read), a PersistenceManager is needed");
		}
		Transaction t = (Transaction) pm.currentTransaction();
		try {
		    SpeedoAccessor sa = (SpeedoAccessor) tpm.writeIntention(t, sp, 
		            isFieldLockingLevel ? thinLock : null);
		    sa.loadFields(pm, fields);
		    return sa;
        } catch (RolledBackPersistenceException e) {
            throw t.rollBackOnInternalError(e);
		} catch (PersistenceException e) {
			throw new JDOException ("Impossible to notify a write intention",
				ExceptionHelper.getNested(e));
		}
	}

	/**
	 * It retrieves the SpeedoAccessor instance used in the current context.
	 * If the proxy is not active then the reference accessor is returned.
	 * If there is an active then the $classNameFields used in the context is 
	 * returned. Be careul, if the persistent object is not used in the current 
	 * context, then a null value will be returned, because no SpeedoAccessor
	 * is registered in the working set.
	 */
	public final SpeedoAccessor getSpeedoAccessor(SpeedoProxy sp) {
		if (!sp.jdoIsActive()) {
			return sp.getReferenceAccessor();
		}
		ProxyManager pm = pmf.lookup();
		if (pm == null) {
		    throw new JDOUserException("When a persistent object is used (read or write), a PersistenceManager is needed");
		}
		return (SpeedoAccessor) ((Transaction) pm.currentTransaction())
			.lookup(sp.getPName());
	}

	public final SpeedoProxy detachCopy(SpeedoProxy sp, 
	        ProxyManager pm, 
	        Map map, 
	        Object clone, 
	        Collection fgHints) {
		Logger logger = getLogger();
		if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
			logger.log(BasicLevel.DEBUG, "detachCopy()");
		}
		if (!isAbstract() && clone == null){
			try {
				//instanciate the SpeedoProxy clone
				clone = newSpeedoProxyInstance(sp.getClass());
				SpeedoProxy spClone = (SpeedoProxy)clone;
				spClone.setEncodedPName(pm.getObjectId(sp));
				spClone.setCeAge(0);
				spClone.jdoIsActive(false);
				//instanciate the clone of the fields
				SpeedoAccessor fieldsClone = sp.createAccessor();
				spClone.setReferenceAccessor(fieldsClone);
				fieldsClone.setSpeedoProxy(spClone);
				//put the association between the speedoproxy and its clone into the map
				if(map != null)
					map.put(sp, clone);
				synchronized(fgHints){
					readIntention(sp, null).detachCopy(pm, map, fieldsClone, fgHints);
				}
				//mark the state as detached
				fieldsClone.setDetachedStatus(DetachedLifeCycle.DETACHED_CLEAN);
			} catch(Exception e){
				throw new JDOUserException("Detach cannot be performed", new Exception[]{ExceptionHelper.getNested(e)});
			}
		}
		return (SpeedoProxy)clone;
	}

	public final void attachCopy(SpeedoProxy sp, 
		ProxyManager pm, 
		Map map, 
		Object clone, 
		SpeedoAccessor sa, 
		boolean makeTransactional) {
		
		if(map != null) {
			if( (SpeedoProxy) map.get(clone) == sp)
				return;
			map.put(clone, sp);
			sa.attachCopy(pm, map, 
				((SpeedoProxy) clone).getReferenceAccessor(), 
				makeTransactional);
		}
	}


	public final Collection fgGetFieldsToLoad(String fgName, boolean onlyDirectRef){
		Collection c = fgGetFieldsToLoad(fgName);
		if(onlyDirectRef){
			Iterator it = c.iterator();
			while(it.hasNext()) {
				String s = (String) it.next();
				if(s.indexOf(SpeedoFetchGroup.FG_DOT) != -1 
				        || s.indexOf(SpeedoFetchGroup.FG_SHARP) != -1 
				        || s.indexOf(SpeedoFetchGroup.FG_SLASH) != -1 
				        || s.indexOf(SpeedoFetchGroup.FG_KEY) != -1
				        || s.indexOf(SpeedoFetchGroup.FG_VALUE) != -1) {
					it.remove();
				}
			}
		}
		return c;
	}
	
    public void setPrefetchOnExtent(boolean prefetch) {
        this.prefetchOnExtent = prefetch;
    }
    

    public boolean getPrefetchOnExtent() {
        return prefetchOnExtent;
    }
    public void setPrefetchOnGenClass(boolean prefetch) {
        this.prefetchOnGenClass = prefetch;
    }
    public boolean getPrefetchOnGenClass() {
        return prefetchOnGenClass;
    }
    public void setPrefetchOnQuery(boolean prefetch) {
        this.prefetchOnQuery = prefetch;
    }
    public boolean getPrefetchOnQuery() {
        return prefetchOnQuery;
    }
	
	public void initSH() {
	}
	
	public Class getClassForQuery(String className, String queryName) {
	    try {
            return getClass().getClassLoader().loadClass(className);
        } catch (ClassNotFoundException e) {
            throw new JDOUserException("Class '" + className 
                    + "' used in the query '" + queryName 
                    + "' defined in by the class '" + getClassName() 
                    + "' is not available. See the inner class loading problem: ", e);
        }
	}
	
	public final QueryDefinition addNamedQuery(String name, QueryDefinition query) {
		return (QueryDefinition) namedQueries.put(name, query);
	}
	public final QueryDefinition removeNamedQuery(String name) {
		return (QueryDefinition) namedQueries.remove(name);
	}
	public final QueryDefinition getNamedQuery(String name) {
		return (QueryDefinition) namedQueries.get(name);
	}
	
	public String getPath() {
	    return getClassName();
	}
	
	//IMPLEMENTATION OF THE UserCacheHelper INTERFACE //
	//------------------------------------------------//
	protected static class UC {
	    public String name;
	    public int id;
	    public UserCache userCache;
	    public String[] fields;
	}
	
	protected UC[] userCaches = new UC[0];

	protected synchronized UC addUserCache(String userCacheName, String[] fields, int id) {
	    UC uc = new UC();
	    uc.name = userCacheName;
	    uc.fields = fields;
	    uc.id = id;
	    uc.userCache = new UserCacheImpl();
	    if (id >= userCaches.length) {
	        UC[] newUserCaches = new UC[id+1];
	        System.arraycopy(userCaches, 0, newUserCaches, 0, userCaches.length);
		    userCaches = newUserCaches;
	    }
	    userCaches[uc.id] = uc;
	    return uc;
	}

	/**
	 * Retrives the name of an user cache from its int identifier.
	 * @param cacheId the identifier of an user cache
	 * @return the name of the user cache, or a null value if the identifier
	 * does not match an existing user cache.
	 */
	public String getUserCacheName(int cacheId) {
	    if (cacheId < 0 || cacheId >= userCaches.length) {
	        return null;
	    }
	    if (userCaches[cacheId] == null) {
	        return null;
	    }
	    return userCaches[cacheId].name; 
	}

	public boolean activeUserCache(String cacheName) {
	    return false;
	}
	public UserCache getUserCache(int cacheId) {
	    if (cacheId < 0 || cacheId >= userCaches.length) {
	        return null;
	    }
	    if (userCaches[cacheId] == null) {
	        return null;
	    }
	    return userCaches[cacheId].userCache; 
	}
	public int[] getUserCacheIds() {
	    int[] ids = new int[userCaches.length];
	    for (int i = 0; i < ids.length; i++) {
		    if (userCaches[i] == null) {
		        ids[i] = -1;
		    } else {
		        ids[i] = userCaches[i].id;
		    }
        }
	    return ids;
	}
	
	public String[] getIndexededFieldsFromUserCache(int cacheId) {
	    if (cacheId < 0 || cacheId >= userCaches.length) {
		    return new String[0];
	    }
	    if (userCaches[cacheId] == null) {
		    return new String[0];
	    }
	    return userCaches[cacheId].fields; 
	}
	
	public int getUserCacheId(Collection fieldNames) {
	    for(int i=0; i<userCaches.length; i++) {
	        if (userCaches[i] == null || userCaches[i].userCache == null) {
	            continue;
	        }
	        String[] fns = userCaches[i].fields;
	        List l = Arrays.asList(fns);
	        if (fns.length == fieldNames.size() 
	                && fieldNames.containsAll(l)
	                && l.containsAll(fieldNames)
	                ) {
	            return userCaches[i].id;
	        }
	    }
	    return -1;
	}
	    
    public Collection getActiveUserCache() {
        if (userCaches == null || userCaches.length == 0) {
            return Collections.EMPTY_LIST;
        }
        ArrayList al = new ArrayList();
	    for(int i=0; i<userCaches.length; i++) {
	        if (userCaches[i] == null || userCaches[i].userCache == null) {
	            continue;
	        }
	        al.add(userCaches[i].userCache);
	    }
        return al;
    }
    public void userCacheEntryUnbound(Object oid) {
        if (userCaches == null || userCaches.length == 0) {
            return;
        }
	    for(int i=0; i<userCaches.length; i++) {
	        if (userCaches[i] == null || userCaches[i].userCache == null) {
	            continue;
	        }
	        userCaches[i].userCache.unbindFromOID(oid);
	    }
    }
}
